﻿//---------------------------------------------------------------------------
//
//   ExportRIB
//
//          Copyright(C) 1999-2015, tetraface Inc.
//
//      Sample of Export plug-in for exporting a RIB (RenderMan) file.
//      Created DLL must be installed in "Plugins\Export" directory.
//
//    　RIB(RenderMan)形式を出力するエクスポートプラグインのサンプル。
//    　作成したDLLは"Plugins\Export"フォルダに入れる必要がある。
//
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#define NOMINMAX
#include <windows.h>
#include <ShellAPI.h>
#include <Shlwapi.h>
#include <ShlObj.h>
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include "MQPlugin.h"
#include "MQBasePlugin.h"
#include "MQWidget.h"
#include "MQSetting.h"
#include "MQ3DLib.h"
#include "FileLib.h"
#include "Language.h"

static wchar_t s_DllPath[_MAX_PATH];

static const wchar_t *intergrator_reyes = L"reyes";
static const wchar_t *intergrator_pathtracer = L"pathtracer";
static const wchar_t *intergrator_vcm = L"vcm";


std::string extractDirectory(const std::string& src_path)
{
	char buf[MAX_PATH];
	strcpy_s(buf, src_path.c_str());
	::PathRemoveFileSpecA(buf);
	return std::string(buf);
}

std::wstring extractDirectory(const std::wstring& src_path)
{
	wchar_t buf[MAX_PATH];
	wcscpy_s(buf, src_path.c_str());
	::PathRemoveFileSpecW(buf);
	return std::wstring(buf);
}

std::string combinePath(const std::string& base_dir, const std::string& cat_dir)
{
	char buf[MAX_PATH];
	::PathCombineA(buf, base_dir.c_str(), cat_dir.c_str());
	return std::string(buf);
}

std::wstring combinePath(const std::wstring& base_dir, const std::wstring& cat_dir)
{
	wchar_t buf[MAX_PATH];
	::PathCombineW(buf, base_dir.c_str(), cat_dir.c_str());
	return std::wstring(buf);
}

std::vector<std::wstring> enumDirectoriesInDirectory(const std::wstring& dir_path, bool exclude_hidden)
{
	std::vector<std::wstring> ret;

	WIN32_FIND_DATAW fd;
	HANDLE handle = ::FindFirstFileW(combinePath(dir_path, L"*").c_str(), &fd);
	if(handle == INVALID_HANDLE_VALUE){
		return ret;
	}

	do {
		if((wcscmp(fd.cFileName, L".") == 0) || (wcscmp(fd.cFileName, L"..") == 0)){
			continue; // current or up directory is ignored
		}

		if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)){
			continue; // file is ignored
		}
		if(exclude_hidden && (fd.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN)){
			continue; // hidden directory is ignored
		}

		ret.push_back(fd.cFileName);
	} while(::FindNextFileW(handle, &fd));

	return ret;
}

std::string WideToAnsi(const wchar_t *ptr)
{
	if(ptr == NULL) return std::string();

	int lenu = WideCharToMultiByte(CP_ACP,0,ptr,-1,NULL,0,NULL,NULL);
	char *stru = (char*)malloc(lenu+1);
	WideCharToMultiByte(CP_ACP,0,ptr,-1,stru,lenu,NULL,NULL);

	std::string str(stru);
	free(stru);
	return str;
}


//---------------------------------------------------------------------------
//  DllMain
//---------------------------------------------------------------------------
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	switch(ul_reason_for_call){
	case DLL_PROCESS_ATTACH:
		::GetModuleFileNameW((HMODULE)hModule, s_DllPath, _MAX_PATH);
		break;
	}

	// Nothing to do and return TRUE.
	//プラグインとしては特に必要な処理はないので、何もせずにTRUEを返す
    return TRUE;
}

class ExportRIBPlugin : public MQExportPlugin
{
public:
	ExportRIBPlugin();
	~ExportRIBPlugin();

	void GetPlugInID(DWORD *Product, DWORD *ID) override;
	const char *GetPlugInName(void) override;
	const char *EnumFileType(int index) override;
	const char *EnumFileExt(int index) override;
	BOOL ExportFile(int index, const char *filename, MQDocument doc) override;

	MLanguage m_Language;

	void LoadResource();

private:
	BOOL SaveRIB(const char *filename, MQDocument doc);
};


MQBasePlugin *GetPluginClass()
{
	static ExportRIBPlugin plugin;
	return &plugin;
}

ExportRIBPlugin::ExportRIBPlugin()
{
}

ExportRIBPlugin::~ExportRIBPlugin()
{
}

//---------------------------------------------------------------------------
//  MQGetPlugInID
//    プラグインIDを返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
void ExportRIBPlugin::GetPlugInID(DWORD *Product, DWORD *ID)
{
	// プロダクト名(制作者名)とIDを、全部で64bitの値として返す
	// 値は他と重複しないようなランダムなもので良い
	*Product = 0x56A31D20;
	*ID      = 0xC486EA1D;
}

//---------------------------------------------------------------------------
//  MQGetPlugInName
//    プラグイン名を返す。
//    この関数は[プラグインについて]表示時に呼び出される。
//---------------------------------------------------------------------------
const char *ExportRIBPlugin::GetPlugInName(void)
{
	// プラグイン名
	return "RIB Exporter   Copyright(C) 1999-2015, tetraface Inc.";
}

//---------------------------------------------------------------------------
//  MQEnumFileType
//    入力または出力可能なファイルタイプを返す。
//    ファイルタイプは別名保存時のダイアログに表示される。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
const char *ExportRIBPlugin::EnumFileType(int index)
{
	switch(index){
	case 0: return "RenderMan (*.rib)";
	}
	return NULL;
}

//---------------------------------------------------------------------------
//  MQEnumFileExt
//    入力または出力可能な拡張子を返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
const char *ExportRIBPlugin::EnumFileExt(int index)
{
	switch(index){
	case 0: return "rib";
	}
	return NULL;
}

//---------------------------------------------------------------------------
//  MQExportFile
//    [別名で保存]で保存するときに呼び出される。
//---------------------------------------------------------------------------
BOOL ExportRIBPlugin::ExportFile(int index, const char *filename, MQDocument doc)
{
	switch(index){
	case 0: return SaveRIB(filename, doc);
	}
	return FALSE;
}

void ExportRIBPlugin::LoadResource()
{
	if(m_Language.GetLanguage().length() == 0){
		std::wstring lang = GetSettingValue(MQSETTINGVALUE_LANGUAGE);
		std::wstring dir = extractDirectory(s_DllPath);
		std::wstring path = combinePath(dir, L"ExportRIB.resource.xml");
		m_Language.Load(lang, path);
	}
}

class MQExportRIBDialog;

struct CreateDialogOptionParam
{
	ExportRIBPlugin *plugin;
	MQExportRIBDialog *dialog;
	
	bool visible_only;
	std::wstring intergrator; // reyes, pathtracer, vcm
	int samples;
	std::wstring env_map;
	int res_x, res_y;
};

class MQExportRIBDialog : public MQDialog
{
public:
	MQExportRIBDialog(int id, CreateDialogOptionParam *param);
	BOOL CheckClick(MQWidgetBase *sender, MQDocument doc);

	MQCheckBox *check_visible;
	MQLabel *label_ris;
	MQComboBox *combo_ris;
	MQLabel *label_sample;
	MQDoubleSpinBox *spin_sample;
	MQLabel *label_envmap;
	MQComboBox *combo_envmap;
	MQLabel *label_res;
	MQSpinBox *spin_res_x;
	MQSpinBox *spin_res_y;

	void UpdateInterface();
private:
	CreateDialogOptionParam *m_Param;
};

MQExportRIBDialog::MQExportRIBDialog(int id, CreateDialogOptionParam *param) : MQDialog(id), m_Param(param)
{
}

void MQExportRIBDialog::UpdateInterface()
{
}

BOOL MQExportRIBDialog::CheckClick(MQWidgetBase *sender, MQDocument doc)
{
	UpdateInterface();
	return FALSE;
}

static void CreateDialogOption(bool init, MQFileDialogCallbackParam *param, void *ptr)
{
	CreateDialogOptionParam *option = (CreateDialogOptionParam*)ptr;
	
	if(init)
	{
		MQExportRIBDialog *dialog = new MQExportRIBDialog(param->dialog_id, option);
		option->dialog = dialog;
		
		MQFrame parent(param->parent_frame_id);	
		MQFrame *hframe;
		MQGroupBox *group = dialog->CreateGroupBox(&parent, option->plugin->m_Language.Search("Option"));
		dialog->check_visible = dialog->CreateCheckBox(group, option->plugin->m_Language.Search("VisibleOnly"));
		hframe = dialog->CreateHorizontalFrame(group);
		dialog->label_ris = dialog->CreateLabel(hframe, option->plugin->m_Language.Search("UseRIS"));
		dialog->combo_ris = dialog->CreateComboBox(hframe);
		dialog->combo_ris->AddItem(option->plugin->m_Language.Search("REYES"));
		dialog->combo_ris->AddItem(option->plugin->m_Language.Search("PathTracer"));
		dialog->combo_ris->AddItem(option->plugin->m_Language.Search("VCM"));
		dialog->combo_ris->SetHintSizeRateX(8.0);
		dialog->combo_ris->SetFillBeforeRate(1.0);
		hframe = dialog->CreateHorizontalFrame(group);
		dialog->label_sample = dialog->CreateLabel(hframe, option->plugin->m_Language.Search("MaxSamples"));
		dialog->spin_sample = dialog->CreateDoubleSpinBox(hframe);
		dialog->spin_sample->SetMin(16);
		dialog->spin_sample->SetMax(1024);
		dialog->spin_sample->SetIncrement(2);
		dialog->spin_sample->SetExponential(true);
		dialog->spin_sample->SetDecimalDigit(0);
		dialog->spin_sample->SetFillBeforeRate(1.0);
		hframe = dialog->CreateHorizontalFrame(group);
		dialog->label_envmap = dialog->CreateLabel(hframe, option->plugin->m_Language.Search("EnvMap"));
		dialog->combo_envmap = dialog->CreateComboBox(hframe);
		dialog->combo_envmap->AddItem(option->plugin->m_Language.Search("None"));
		dialog->combo_envmap->AddItem(L"desert");
		dialog->combo_envmap->AddItem(L"street");
		dialog->combo_envmap->SetFillBeforeRate(1.0);
		hframe = dialog->CreateHorizontalFrame(group);
		dialog->label_res = dialog->CreateLabel(hframe, option->plugin->m_Language.Search("Resolution"));
		dialog->spin_res_x = dialog->CreateSpinBox(hframe);
		dialog->spin_res_x->SetMin(16);
		dialog->spin_res_x->SetMax(1280);
		dialog->label_res = dialog->CreateLabel(hframe, L"x");
		dialog->spin_res_y = dialog->CreateSpinBox(hframe); 
		dialog->spin_res_y->SetMin(16);
		dialog->spin_res_y->SetMax(960);

		dialog->check_visible->SetChecked(option->visible_only);
		if(option->intergrator == intergrator_reyes) dialog->combo_ris->SetCurrentIndex(0);
		else if(option->intergrator == intergrator_vcm) dialog->combo_ris->SetCurrentIndex(2);
		else dialog->combo_ris->SetCurrentIndex(1);
		dialog->spin_sample->SetPosition((double)option->samples);
		if(option->env_map == L"desert") dialog->combo_envmap->SetCurrentIndex(1);
		else if(option->env_map == L"street") dialog->combo_envmap->SetCurrentIndex(2);
		else dialog->combo_envmap->SetCurrentIndex(0);
		dialog->spin_res_x->SetPosition(option->res_x);
		dialog->spin_res_y->SetPosition(option->res_y);

		dialog->UpdateInterface();
	}
	else
	{
		MQExportRIBDialog *dialog = option->dialog;
		option->visible_only = dialog->check_visible->GetChecked();
		switch(dialog->combo_ris->GetCurrentIndex()){
		case 0: option->intergrator = intergrator_reyes; break;
		default:option->intergrator = intergrator_pathtracer; break;
		case 2: option->intergrator = intergrator_vcm; break;
		}
		option->samples = (int)dialog->spin_sample->GetPosition();
		switch(dialog->combo_envmap->GetCurrentIndex()){
		case 1: option->env_map = L"desert"; break;
		case 2: option->env_map = L"street"; break;
		default: option->env_map.clear(); break;
		}
		option->res_x = dialog->spin_res_x->GetPosition();
		option->res_y = dialog->spin_res_y->GetPosition();
		delete option->dialog;
	}
}

void MatrixMultiply3x3(MQMatrix& dst, const MQMatrix& m1, const MQMatrix& m2)
{
	MQMatrix t1(m1);
	MQMatrix t2(m2);

	dst._11 = t1._11 * t2._11 + t1._12 * t2._21 + t1._13 * t2._31;
	dst._12 = t1._11 * t2._12 + t1._12 * t2._22 + t1._13 * t2._32;
	dst._13 = t1._11 * t2._13 + t1._12 * t2._23 + t1._13 * t2._33;
	dst._21 = t1._21 * t2._11 + t1._22 * t2._21 + t1._23 * t2._31;
	dst._22 = t1._21 * t2._12 + t1._22 * t2._22 + t1._23 * t2._32;
	dst._23 = t1._21 * t2._13 + t1._22 * t2._23 + t1._23 * t2._33;
	dst._31 = t1._31 * t2._11 + t1._32 * t2._21 + t1._33 * t2._31;
	dst._32 = t1._31 * t2._12 + t1._32 * t2._22 + t1._33 * t2._32;
	dst._33 = t1._31 * t2._13 + t1._32 * t2._23 + t1._33 * t2._33;
}

void MatrixRotateZ(MQMatrix& mtx, float radian)
{
	MQMatrix z_rotate;
	z_rotate.Identify();
	z_rotate._11 = z_rotate._22 = cos(radian);
	z_rotate._12 = sin(radian);
	z_rotate._21 = -z_rotate._12;
	MatrixMultiply3x3(mtx, mtx, z_rotate);
}

MQMatrix GetCameraMatrixFromAngle(MQAngle angle)
{
	float ch = cos(RAD(angle.head));
	float sh = sin(RAD(angle.head));
	float cp = cos(RAD(angle.pich));
	float sp = sin(RAD(angle.pich));
	
	MQMatrix mtx;
	mtx.Identify();
	mtx._11 = ch;
	mtx._12 = sh*sp;
	mtx._13 = sh*cp;
	mtx._21 = 0;
	mtx._22 = cp;
	mtx._23 = -sp;
	mtx._31 = sh;
	mtx._32 = -ch*sp;
	mtx._33 = -ch*cp;
	if(angle.bank != 0) MatrixRotateZ(mtx, RAD(angle.bank));
	return mtx;
}

class MQRenderingWindow : public MQWindow
{
public:
	// Constructor
	MQRenderingWindow(MQWindowBase& parent, ExportRIBPlugin *plugin);
	// Destructor
	~MQRenderingWindow();

private:
	MQLabel *Label;

};

// Constructor
MQRenderingWindow::MQRenderingWindow(MQWindowBase& parent, ExportRIBPlugin *plugin)
	: MQWindow(parent)
{
	this->SetCloseButton(false);
	this->SetMaximizeButton(false);
	this->SetMinimizeButton(false);
	this->SetCanResize(false);
	//this->SetTitle(plugin->GetResourceString("Progress"));
	this->SetOutSpace(0.2);

	Label = CreateLabel(this);
	Label->SetWordWrap(true);
	Label->SetText(plugin->m_Language.Search("NowRendering"));
	Label->SetHintSizeRateX(20.0);
}

MQRenderingWindow::~MQRenderingWindow()
{
}


// RIB形式で出力
BOOL ExportRIBPlugin::SaveRIB(const char *filename, MQDocument doc)
{
	int i, face, mat_index;
	MQMaterial mat;
	char path[MAX_PATH], name[64];

	LoadResource();

	CreateDialogOptionParam option;
	option.plugin = this;
	option.visible_only = true;
	option.intergrator = L"pathtracer";
	option.samples = 256;
	option.res_x = 512;
	option.res_y = 512;

	MQSetting *setting = OpenSetting();
	if(setting != NULL){
		setting->Load("VisibleOnly", option.visible_only, option.visible_only);
		setting->Load("Integrator", option.intergrator, option.intergrator);
		setting->Load("MaxSamples", option.samples, option.samples);
		setting->Load("EnvMap", option.env_map, option.env_map);
		setting->Load("ResolutionX", option.res_x, option.res_x);
		setting->Load("ResolutionY", option.res_y, option.res_y);
		CloseSetting(setting);
	}

	// 座標軸変換用ダイアログの表示
	MQFileDialogInfo dlginfo;
	memset(&dlginfo, 0, sizeof(dlginfo));
	dlginfo.dwSize = sizeof(dlginfo);
	dlginfo.axis_x = MQFILE_TYPE_RIGHT;
	dlginfo.axis_y = MQFILE_TYPE_UP;
	dlginfo.axis_z = MQFILE_TYPE_BACK;
	dlginfo.softname = "RenderMan";
	dlginfo.dialog_callback = CreateDialogOption;
	dlginfo.dialog_callback_ptr = &option;
	MQ_ShowFileDialog("RIB Export", &dlginfo);

	// Save the settings.
	setting = OpenSetting();
	if(setting != NULL){
		setting->Save("VisibleOnly", option.visible_only);
		setting->Save("Integrator", option.intergrator);
		setting->Save("MaxSamples", option.samples);
		setting->Save("EnvMap", option.env_map);
		setting->Save("ResolutionX", option.res_x);
		setting->Save("ResolutionY", option.res_y);
		CloseSetting(setting);
	}

	// ファイルを作成・オープンする
	FILE *fh;
	errno_t ret_s = fopen_s(&fh, filename, "w");
	if(ret_s != 0)
		return FALSE;

	MQWaitCursorChanger wait_cursor(this);

	MQScene scene = doc->GetScene(0);

	// 複数のオブジェクトを１つに合成するために、作業用オブジェクトを作成する
	MQObject obj = MQ_CreateObject();

	// 作業用オブジェクトにすべてのオブジェクトを合成
	int objcount = doc->GetObjectCount();
	for(i=0; i<objcount; i++)
	{
		MQObject thisobj = doc->GetObject(i);
		if(thisobj == NULL) 
			continue;
		if(option.visible_only){
			if(thisobj->GetVisible() == 0)
				continue;
		}
		// フリーズなどにより内容が変更されてしまうので、オブジェクトを複製しておく
		thisobj = thisobj->Clone();
		// 曲面などの属性をフリーズ
		thisobj->Freeze(MQOBJECT_FREEZE_ALL);
		// 作業用オブジェクトに合成する
		obj->Merge(thisobj);
		// 複製したオブジェクトはもう不要なので破棄
		thisobj->DeleteThis();
	}

	// 面と頂点の数を取得
	int face_count = obj->GetFaceCount();
	int vert_count = obj->GetVertexCount();

	// 座標軸をダイアログで指定したものに変換
	for(i=0; i<vert_count; i++){
		if(obj->GetVertexRefCount(i) == 0) continue;
		MQPoint p = obj->GetVertex(i);
		MQ_ExportAxis(&dlginfo, &p, 1);
		obj->SetVertex(i, p);
	}

	MQObjNormal normal(obj);

	// ファイルの先頭にデータを書き込み
	fprintf(fh, "##RenderMan RIB\n");
	fprintf(fh, "\n");

	fprintf(fh, "Option \"searchpath\" \"shader\" [\".:&\"]\n");
	fprintf(fh, "Format %d %d 1\n", option.res_x, option.res_y);
	GetFileNameFromPath(path, filename);
	fprintf(fh, "Display \"%s.tif\" \"file\" \"rgba\"\n", path);
	if(option.intergrator == intergrator_pathtracer){
		fprintf(fh, "Integrator \"PxrPathTracer\" \"handle\" \"int numLightSamples\" [4] \"int numBxdfSamples\" [4]\n");
		fprintf(fh, "Hider \"raytrace\" \"constant string integrationmode\" [\"path\"] \"constant int incremental\" [0] \"int maxsamples\" [%d]\n", option.samples);
	}else if(option.intergrator == intergrator_vcm){
		fprintf(fh, "Integrator \"PxrVCM\" \"handle\" \"int maxPathLength\" [13] \"int mergePaths\" [1]\n");
		fprintf(fh, "Hider \"raytrace\" \"int minsamples\" [64] \"int maxsamples\" [%d] \"int incremental\" [1]\n", option.samples);
	}
	float scene_scale = 1.0f;
	if(scene->GetOrtho()){
		scene_scale = 0.001f * scene->GetZoom();
		fprintf(fh, "Projection \"orthographic\" \n");
	}else{
		fprintf(fh, "Projection \"perspective\" \"fov\" %.1f\n", scene->GetFOV()/3.14159f*180.0f);
	}
	MQAngle angle = scene->GetCameraAngle();
	MQMatrix cammtx = GetCameraMatrixFromAngle(angle);
	MQPoint campos = scene->GetCameraPosition() * cammtx * scene_scale;
	fprintf(fh, "Translate %f %f %f\n", -campos.x, -campos.y, -campos.z);
	fprintf(fh, "Rotate %f 0 0 1\n", angle.bank);
	fprintf(fh, "Rotate %f 1 0 0\n", -angle.pitch);
	fprintf(fh, "Rotate %f 0 1 0\n", -angle.head);
	if(scene_scale != 1.0f){
		fprintf(fh, "Scale %f %f %f\n", scene_scale, scene_scale, scene_scale);
	}
	fprintf(fh, "\n");

	MQColor amb_color = scene->GetGlobalAmbientColor();
	fprintf(fh, "WorldBegin\n");
	fprintf(fh, "\n");

	bool use_ris = (option.intergrator != intergrator_reyes);

	if(use_ris){
		// Distant lights
		int light_num = scene->GetGlobalDirectionalLightNum();
		for(i=0; i<light_num; i++){
			MQPoint axisX, axisY, axisZ;
			axisZ = -scene->GetGlobalDirectionalLightDirection(i);
			MQ_ExportAxis(&dlginfo, &axisZ, 1);
			axisZ.normalize();
			if(fabs(GetInnerProduct(axisZ, MQPoint(0,0,1))) < fabs(GetInnerProduct(axisZ, MQPoint(0,1,0)))){
				axisX = GetCrossProduct(axisZ, MQPoint(0,0,1));
				axisY = GetCrossProduct(axisX, axisZ);
			}else{
				axisX = GetCrossProduct(axisZ, MQPoint(0,1,0));
				axisY = GetCrossProduct(axisX, axisZ);
			}
			fprintf(fh, "AttributeBegin\n");
			fprintf(fh, "Transform [%.5f %.5f %.5f 0\n", axisX.x, axisX.y, axisX.z);
			fprintf(fh, "	%.5f %.5f %.5f 0\n", axisY.x, axisY.y, axisY.z);
			fprintf(fh, "	%.5f %.5f %.5f 0\n", axisZ.x, axisZ.y, axisZ.z);
			fprintf(fh, "	0 0 0 1]\n");
			fprintf(fh, "AreaLightSource \"PxrAreaLight\" \"light%d\" "
				"\"string shape\" [\"distant\"] \"float intensity\" [0.5]\n", 
				i+1);
			fprintf(fh, "ShadingRate 1000\n");
			fprintf(fh, "Geometry \"distantlight\"\n");
			fprintf(fh, "AttributeEnd\n");
			fprintf(fh, "\n");
		}

		// Environment map light
		if(option.env_map.length() > 0){
			fprintf(fh, "AttributeBegin\n");
			fprintf(fh, "AreaLightSource \"PxrEnvMapLight\" \"envlight\" \"float intensity\" [0.5] \"string envmap\" [\"%s.env\"]\n", WideToAnsi(option.env_map.c_str()).c_str());
			fprintf(fh, "Bxdf \"PxrLightEmission\" \"envlight\"\n");
			fprintf(fh, "Sides 1\n");
			fprintf(fh, "Opacity [1 1 1]\n");
			fprintf(fh, "Rotate -90 1 0 0\n");
			fprintf(fh, "Geometry \"envsphere\"\n");
			fprintf(fh, "AttributeEnd\n");
			fprintf(fh, "\n");
		}
	}else{
		// Distant lights
		fprintf(fh, "LightSource \"ambientlight\" 1 \"lightcolor\" [%.3f %.3f %.3f]\n", amb_color.r, amb_color.g, amb_color.b);
		fprintf(fh, "\n");
		fprintf(fh, "Declare \"shadows\" \"string\"\n");
		fprintf(fh, "Attribute \"light\" \"shadows\" \"on\"\n");
		int light_num = scene->GetGlobalDirectionalLightNum();
		for(i=0; i<light_num; i++){
			MQPoint dir = scene->GetGlobalDirectionalLightDirection(i);
			MQ_ExportAxis(&dlginfo, &dir, 1);
			dir.normalize();
			MQColor col = scene->GetGlobalDirectionalLightColor(i);
			fprintf(fh, "LightSource \"distantlight\" %d \"from\" [%f %f %f] \"to\" [0 0 0] \"lightcolor\" [%.3f %.3f %.3f]\n", 
				i+2, -dir.x, -dir.y, -dir.z, col.r, col.g, col.b);
		}
		fprintf(fh, "\n");
	}

	// それぞれの材質がいくつの面に使われているかどうかを調べておく
	int mat_count = doc->GetMaterialCount();
	std::vector<int> mat_used(mat_count+1);
	for(i=0; i<=mat_count; i++)
		mat_used[i] = 0;
	for(face=0; face<face_count; face++)
	{
		if(obj->GetFacePointCount(face) < 3) continue;
		mat_index = obj->GetFaceMaterial(face);
		mat_used[(mat_index>=0) ? mat_index : mat_count]++; // 未着色面の場合は配列の最後にデータをセット
	}

	std::set<std::string> converted_textures;
	std::string file_dir = extractDirectory(filename);

	// 材質ごとにAttributeとしてポリゴン情報を記録する
	for(mat_index=0; mat_index<=mat_count; mat_index++)
	{
		int mi;

		// この材質が使われていなければパス
		if(mat_used[mat_index] == 0)
			continue;

		if(mat_index == mat_count) {
			mat = MQ_CreateMaterial(); // 未着色面の場合は一時的に材質を作成
			mi = -1;
		} else {
			mat = doc->GetMaterial(mat_index);
			mi = mat_index;
		}
	
		fprintf(fh, "AttributeBegin\n");

		// コメントとして材質名を記録しておく
		mat->GetName(name, sizeof(name));
		fprintf(fh, "  # %s\n", name);

		MQShaderNode shader_node = mat->GetMainShaderNode();
		bool enable_uv = false;

		if(use_ris){
			std::string shader_rib_name;
			if(shader_node != NULL)
				shader_rib_name = shader_node->GetOptionValue("rib");
			if(!shader_rib_name.empty()){
				// For a RenderMan shader

				// Find a texture.
				char texname[MAX_PATH];
				texname[0] = '\0';
				shader_node->GetMappingFilename("Texture", texname, MAX_PATH);

				if(texname[0] != '\0'){
					enable_uv = true;
					fprintf(fh, "  Pattern \"PxrTexture\" \"%s_texture\" \"string filename\" [\"%s\"] \"int invertT\" [0]\n", name, texname);
				}

				// Presence(alpha) pattern
				int param_num = shader_node->GetParameterNum();
				for(int pi = 0; pi < param_num; pi++){
					std::string param_name = shader_node->GetParameterName(pi);
					std::string param_rib_name = shader_node->GetParameterOptionValue(pi, "rib_name");
					if(param_name == "Alpha" && param_rib_name == "presence"){
						float alpha = shader_node->GetParameterFloatValue(pi, 0);
						if(alpha < 1){
							fprintf(fh, "  Pattern \"PxrToFloat\" \"%s_presence\" \"color input\" [%.3g %.3g %.3g]\n", name, alpha, alpha, alpha);
						}
						break;
					}
				}

				// Bxdf shader
				fprintf(fh, "  Bxdf \"%s\" \"%s\"", shader_rib_name.c_str(), name);

				for(int pi = 0; pi < param_num; pi++){
					std::string param_rib_name = shader_node->GetParameterOptionValue(pi, "rib_name");
					if(param_rib_name.empty()) continue;

					std::string param_name = shader_node->GetParameterName(pi);
					if(param_name == "BaseColor"){
						if(texname[0] != '\0'){
							fprintf(fh, " \"reference color %s\" [\"%s_texture:resultRGB\"]", param_rib_name.c_str(), name);
						}else{
							MQColorRGBA color = shader_node->GetParameterColorValue(pi, 0);
							fprintf(fh, " \"color %s\" [%.3g %.3g %.3g]", param_rib_name.c_str(), color.r, color.g, color.b);
						}
					}else if(param_name == "Alpha" && param_rib_name == "presence"){
						float alpha = shader_node->GetParameterFloatValue(pi, 0);
						if(alpha < 1){
							fprintf(fh, " \"reference float %s\" [\"%s_presence:resultF\"]", param_rib_name.c_str(), name);
						}else if(texname[0] != '\0'){
							fprintf(fh, " \"reference float %s\" [\"%s_texture:resultA\"]", param_rib_name.c_str(), name);
						}
					}else{
						switch(shader_node->GetParameterValueType(pi)){
						case MQMATERIAL_SHADER_PARAM_TYPE_BOOL:
							fprintf(fh, " \"int %s\" [%d]", param_rib_name.c_str(), shader_node->GetParameterBoolValue(pi, 0) ? 1 : 0);
							break;
						case MQMATERIAL_SHADER_PARAM_TYPE_INT:
							fprintf(fh, " \"int %s\" [%d]", param_rib_name.c_str(), shader_node->GetParameterIntValue(pi, 0));
							break;
						case MQMATERIAL_SHADER_PARAM_TYPE_FLOAT:
							fprintf(fh, " \"float %s\" [%.6g]", param_rib_name.c_str(), shader_node->GetParameterFloatValue(pi, 0));
							break;
						case MQMATERIAL_SHADER_PARAM_TYPE_COLOR:
							MQColorRGBA col = shader_node->GetParameterColorValue(pi, 0);
							fprintf(fh, " \"color %s\" [%.6g %.6g %.6g]", param_rib_name.c_str(), col.r, col.g, col.b);
							break;
						}
					}
				}
				fprintf(fh, "\n");
			}
			else{
				// For a standard shader
				MQColor color = mat->GetColor();
				float alpha = mat->GetAlpha();
				MQColor emission = mat->GetEmissionColor();

				char texname[MAX_PATH];
				mat->GetTextureName(texname, MAX_PATH);

				if(texname[0] != '\0'){
					enable_uv = true;
				}

				if(texname[0] != '\0'){
					fprintf(fh, "  Pattern \"PxrTexture\" \"%s_texture\" \"string filename\" [\"%s\"] \"int invertT\" [0]\n", name, texname);
				}

				// Presence(alpha) pattern
				if(alpha < 1){
					fprintf(fh, "  Pattern \"PxrToFloat\" \"%s_presence\" \"color input\" [%.3g %.3g %.3g]\n", name, alpha, alpha, alpha);
				}

				// Disney shader
				fprintf(fh, "  Bxdf \"PxrDisney\" \"%s\"", name);
				if(texname[0] != '\0'){
					fprintf(fh, " \"reference color baseColor\" [\"%s_texture:resultRGB\"]", name);
				}else{
					fprintf(fh, " \"color baseColor\" [%.3g %.3g %.3g]", color.r, color.g, color.b);
				}
				if(alpha < 1){
					fprintf(fh, " \"reference float presence\" [\"%s_presence:resultF\"]", name);
				}else if(texname[0] != '\0'){
					fprintf(fh, " \"reference float presence\" [\"%s_texture:resultA\"]", name);
				}
				fprintf(fh, " \"color emitColor\" [%.3g %.3g %.3g]", emission.r, emission.g, emission.b);
				fprintf(fh, " \"float metallic\" [%.3g]", mat->GetReflection());
				fprintf(fh, " \"float specular\" [%.3g]", mat->GetSpecular());
				fprintf(fh, "\n");
			}

		}else{
			MQColor color = mat->GetColor();
			float alpha = mat->GetAlpha();

			char texname[MAX_PATH];
			mat->GetTextureName(texname, MAX_PATH);

			if(texname[0] != '\0'){
				enable_uv = true;
			}

			fprintf(fh, "  Color [ %.3f %.3f %.3f ]\n", color.r, color.g, color.b);
			fprintf(fh, "  Opacity [ %.3f %.3f %.3f ]\n", alpha, alpha, alpha);

			switch(mat->GetShader()){
			case MQMATERIAL_SHADER_CONSTANT:
				fprintf(fh, "  Surface \"constant\"\n");
				break;
			case MQMATERIAL_SHADER_LAMBERT:
				fprintf(fh, "  Surface \"matte\" \"Ka\" %.3f \"Kd\" %.3f\n", mat->GetAmbient(), mat->GetDiffuse());
				break;
			default:
				{
					MQColor spc_col = mat->GetSpecularColor();
					if(texname[0] != '\0'){
						fprintf(fh, "  Surface \"paintedplastic\" \"Ka\" %.3f \"Kd\" %.3f \"Ks\" 1 \"roughness\" %.3f \"specularcolor\" [%.3f %.3f %.3f] \"texturename\" \"%s\"\n", 
							mat->GetAmbient(), mat->GetDiffuse(), mat->GetPower(), spc_col.r, spc_col.g, spc_col.b, texname);
					}else{
						fprintf(fh, "  Surface \"plastic\" \"Ka\" %.3f \"Kd\" %.3f \"Ks\" 1 \"roughness\" %.3f \"specularcolor\" [%.3f %.3f %.3f]\n", 
							mat->GetAmbient(), mat->GetDiffuse(), mat->GetPower(), spc_col.r, spc_col.g, spc_col.b);
					}
				}
				break;
			}
		}

		for(face=0; face<face_count; face++)
		{
			int vert, count;

			// 面中の頂点数を取得
			count = obj->GetFacePointCount(face);
			if(count < 3) continue;

			// この面に割り当てられた材質を調べる
			if(obj->GetFaceMaterial(face) != mi)
				continue;

			// 面中の頂点インデックスを取得
			std::vector<int> vert_index(count);
			obj->GetFacePointArray(face, &(*vert_index.begin()));

			// Position
			fprintf(fh, "  Polygon \"P\" [ ");
			for(vert=0; vert<count; vert++){
				MQPoint p = obj->GetVertex(vert_index[vert]);
				fprintf(fh, "%.6g %.6g %.6g ", p.x, p.y, p.z);
			}
			fprintf(fh, "] ");
			// Normal
			fprintf(fh, "\"N\" [ ");
			for(vert=0; vert<count; vert++){
				MQPoint p = normal.Get(face, vert);
				fprintf(fh, "%.6g %.6g %.6g ", -p.x, -p.y, -p.z);
			}
			fprintf(fh, "] ");
			if(enable_uv){
				// Texture coordinates
				fprintf(fh, "\"st\" [ ");
				std::vector<MQCoordinate> uvarray(count);
				obj->GetFaceCoordinateArray(face, &(*uvarray.begin()));
				for(vert=0; vert<count; vert++){
					fprintf(fh, "%.5g %.5g ", uvarray[vert].u, uvarray[vert].v);
				}
				fprintf(fh, "] ");
			}
			fprintf(fh, "\n");
		}

		fprintf(fh, "AttributeEnd\n");
		fprintf(fh, "\n");

		// 未着色面用に一時的に作った材質を削除
		if(mat_index == mat_count)
			mat->DeleteThis();
	}

	// ファイルの終端のデータを書き込み
	fprintf(fh, "WorldEnd\n");
	fprintf(fh, "\n");

	// ファイルをクローズ
	fclose(fh);

	// 不要になったデータを破棄
	mat_used.clear();
	obj->DeleteThis();

	return TRUE;
}

